@@ -2,20 +2,23 @@ import docker
 import logging
 import json
 import time
+import ipaddress
 
 from json.decoder import JSONDecodeError
 from typing import List, Union, Dict
 from docker.models.containers import Container
 
 from vantage6.common.globals import APPNAME, VPN_CONFIG_FILE
+from vantage6.common.docker.addons import (
+    remove_container_if_exists, remove_container
+)
 from vantage6.node.util import logger_name
 from vantage6.node.globals import (
     MAX_CHECK_VPN_ATTEMPTS, NETWORK_CONFIG_IMAGE, VPN_CLIENT_IMAGE,
-    FREE_PORT_RANGE, DEFAULT_ALGO_VPN_PORT
+    FREE_PORT_RANGE, DEFAULT_ALGO_VPN_PORT, ALPINE_IMAGE
 )
-from vantage6.node.docker.network_manager import IsolatedNetworkManager
+from vantage6.common.docker.network_manager import NetworkManager
 from vantage6.node.docker.docker_base import DockerBaseManager
-from vantage6.node.docker.utils import remove_container
 
 
 class VPNManager(DockerBaseManager):
@@ -25,13 +28,43 @@ class VPNManager(DockerBaseManager):
     """
     log = logging.getLogger(logger_name(__name__))
 
-    def __init__(self, isolated_network_mgr: IsolatedNetworkManager,
-                 node_name: str, vpn_volume_name: str, vpn_subnet: str):
+    def __init__(self, isolated_network_mgr: NetworkManager,
+                 node_name: str, vpn_volume_name: str, vpn_subnet: str,
+                 alpine_image: Union[str, None] = None,
+                 vpn_client_image: Union[str, None] = None,
+                 network_config_image: Union[str, None] = None) -> None:
+        """
+        Initializes a VPN manager instance
+
+        Parameters
+        ----------
+        isolated_network_mgr: NetworkManager
+            An object that manages the node's isolated network
+        node_name: str
+            The name of the node (from config)
+        vpn_volume_name: str
+            The name of the volume in which the VPN data resides
+        vpn_subnet: str
+            The IP mask of the VPN subnet
+        alpine_image: str or None
+            Name of alternative Alpine image to be used
+        """
         super().__init__(isolated_network_mgr)
 
         self.vpn_client_container_name = f'{APPNAME}-{node_name}-vpn-client'
         self.vpn_volume_name = vpn_volume_name
         self.subnet = vpn_subnet
+        self.alpine_image = ALPINE_IMAGE if not alpine_image \
+            else alpine_image
+        self.vpn_client_image = VPN_CLIENT_IMAGE if not vpn_client_image else \
+            vpn_client_image
+        self.network_config_image = NETWORK_CONFIG_IMAGE \
+            if not network_config_image else network_config_image
+
+        self.log.debug('Used VPN images:')
+        self.log.debug(f'  Alpine: {self.alpine_image}')
+        self.log.debug(f'  Client: {self.vpn_client_image}')
+        self.log.debug(f'  Config: {self.network_config_image}')
 
         self.has_vpn = False
 
@@ -40,6 +73,16 @@ class VPNManager(DockerBaseManager):
         Start VPN client container and configure network to allow
         algorithm-to-algoritm communication
         """
+        if not self.subnet:
+            self.log.warn("VPN subnet is not defined! Disabling VPN...")
+            self.log.info("Define the 'vpn_subnet' field in your configuration"
+                          " if you want to use VPN")
+            return
+        elif not self._is_ipv4_subnet(self.subnet):
+            self.log.error(f"VPN subnet {self.subnet} is not a valid subnet! "
+                           "Disabling VPN...")
+            return
+
         self.log.debug("Mounting VPN configuration file")
         # add volume containing OVPN config file
         data_path = '/mnt/vpn/'  # TODO obtain from DockerNodeContext
@@ -51,12 +94,14 @@ class VPNManager(DockerBaseManager):
         env = {'VPN_CONFIG': vpn_config}
 
         # if a VPN container is already running, kill and remove it
-        self.remove_container_if_exists(name=self.vpn_client_container_name)
+        remove_container_if_exists(
+            docker_client=self.docker, name=self.vpn_client_container_name
+        )
 
         # start vpnclient
         self.log.debug("Starting VPN client container")
         self.vpn_client_container = self.docker.containers.run(
-            image=VPN_CLIENT_IMAGE,
+            image=self.vpn_client_image,
             command="",  # commands to run are already defined in docker image
             volumes=volumes,
             detach=True,
@@ -74,41 +119,61 @@ class VPNManager(DockerBaseManager):
             aliases=[self.vpn_client_container_name]
         )
 
+        # check successful initiation of VPN connection
+        if self.has_connection():
+            self.log.info("VPN client container was successfully started!")
+        else:
+            raise ConnectionError("VPN connection not established!")
+
+        # check that the VPN connection IP address is part of the subnet
+        # defined in the node configuration. If not, the VPN connection would
+        # not work.
+        if not self._vpn_in_right_subnet():
+            self.log.error(
+                "The VPN subnet defined in the node configuration file does "
+                "not match the VPN server subnet. Turning off VPN..."
+            )
+            self.exit_vpn(cleanup_host_rules=False)
+            return
+
         # create network exception so that packet transfer between VPN network
         # and the vpn client container is allowed
-        bridge_interface = self._find_isolated_bridge()
-        if not bridge_interface:
+        self.isolated_bridge = self._find_isolated_bridge()
+        if not self.isolated_bridge:
             self.log.error("Setting up VPN failed: could not find bridge "
                            "interface of isolated network")
             return
-        self._configure_host_network(isolated_bridge=bridge_interface)
-
-        # set successful initiation of VPN connection
-        self.has_vpn = True
-        self.log.debug("VPN client container was started")
+        self._configure_host_network()
 
     def has_connection(self) -> bool:
         """ Return True if VPN connection is active """
-        if not self.has_vpn:
-            return False
-        # check if the VPN container has an IP address in the VPN namespace
-        try:
-            # if there is a VPN connection, the following command will return
-            # a json vpn interface. If not, it will return "Device "tun0" does
-            # not exist."
-            _, vpn_interface = self.vpn_client_container.exec_run(
-                'ip --json addr show dev tun0'
-            )
-            vpn_interface = json.loads(vpn_interface)
-        except JSONDecodeError:
-            self.has_vpn = False
-            return False
-        self.has_vpn = True  # TODO rid boolean and only use this function?
-        return True
+        self.log.debug("Waiting for VPN connection. This may take a minute...")
+        n_attempt = 0
+        self.has_vpn = False
+        while n_attempt < MAX_CHECK_VPN_ATTEMPTS:
+            n_attempt += 1
+            try:
+                _, vpn_interface = self.vpn_client_container.exec_run(
+                    'ip --json addr show dev tun0'
+                )
+                vpn_interface = json.loads(vpn_interface)
+                self.has_vpn = True
+                break
+            except (JSONDecodeError, docker.errors.APIError):
+                # JSONDecodeError if VPN is not setup yet, APIError if VPN
+                # container is restarting (e.g. due to connection errors)
+                time.sleep(1)
+        return self.has_vpn
 
-    def exit_vpn(self) -> None:
+    def exit_vpn(self, cleanup_host_rules: bool = True) -> None:
         """
         Gracefully shutdown the VPN and clean up
+
+        Parameters
+        ----------
+        cleanup_host_rules: bool, optional
+            Whether or not to clear host configuration rules. Should be True
+            if they have been created at the time this function runs.
         """
         if not self.has_vpn:
             return
@@ -117,17 +182,24 @@ class VPNManager(DockerBaseManager):
         remove_container(self.vpn_client_container, kill=True)
 
         # Clean up host network changes. We have added two rules to the front
-        # of the DOCKER-USER chain. Now we remove the first two rules (which is
-        # done by removing rule number 1 twice)
-        command = \
-            'sh -c "iptables -D DOCKER-USER 1; iptables -D DOCKER-USER 1;"'
-        self.docker.containers.run(
-            image=NETWORK_CONFIG_IMAGE,
-            network='host',
-            cap_add='NET_ADMIN',
-            command=command,
-            remove=True,
-        )
+        # of the DOCKER-USER chain. We now execute more or less the same
+        # commands, but with -D (delete) instead of -I (insert)
+        if cleanup_host_rules:
+            command = (
+                'sh -c "'
+                f'iptables -D DOCKER-USER -d {self.subnet} '
+                f'-i {self.isolated_bridge} -j ACCEPT; '
+                f'iptables -D DOCKER-USER -s {self.subnet} '
+                f'-o {self.isolated_bridge} -j ACCEPT; '
+                '"'
+            )
+            self.docker.containers.run(
+                image=self.network_config_image,
+                network='host',
+                cap_add='NET_ADMIN',
+                command=command,
+                remove=True,
+            )
 
     def get_vpn_ip(self) -> str:
         """
@@ -138,25 +210,20 @@ class VPNManager(DockerBaseManager):
         str
             IP address assigned to VPN client container by VPN server
         """
-        # VPN might not be fully set up at this point. Therefore, poll to
-        # check. When it is ready, extract the IP address.
-        n_attempt = 0
-        while n_attempt < MAX_CHECK_VPN_ATTEMPTS:
-            n_attempt += 1
-            try:
-                _, vpn_interface = self.vpn_client_container.exec_run(
-                    'ip --json addr show dev tun0'
-                )
-                vpn_interface = json.loads(vpn_interface)
-                break
-            except (JSONDecodeError, docker.errors.APIError):
-                # JSONDecodeError if VPN is not setup yet, APIError if VPN
-                # container is restarting (e.g. due to connection errors)
-                time.sleep(1)
+        try:
+            _, vpn_interface = self.vpn_client_container.exec_run(
+                'ip --json addr show dev tun0'
+            )
+            vpn_interface = json.loads(vpn_interface)
+        except (JSONDecodeError, docker.errors.APIError):
+            # JSONDecodeError if VPN is not setup yet, APIError if VPN
+            # container is restarting (e.g. due to connection errors)
+            raise ConnectionError(
+                "Could not get VPN IP: VPN is not connected!")
         return vpn_interface[0]['addr_info'][0]['local']
 
     def forward_vpn_traffic(self, helper_container: Container,
-                            algo_image_name: str) -> int:
+                            algo_image_name: str) -> List[Dict]:
         """
         Setup rules so that traffic is properly forwarded between the VPN
         container and the algorithm container (and its helper container)
@@ -170,13 +237,14 @@ class VPNManager(DockerBaseManager):
 
         Returns
         -------
-        int
-            Port on the VPN client that forwards traffic to the algo container
+        List[Dict] or None
+            Description of each port on the VPN client that forwards traffic to
+            the algo container. None if VPN is not set up.
         """
-        vpn_port = self._forward_traffic_to_algorithm(
+        ports = self._forward_traffic_to_algorithm(
             helper_container, algo_image_name)
         self._forward_traffic_from_algorithm(helper_container)
-        return vpn_port
+        return ports
 
     def _forward_traffic_from_algorithm(
             self, algo_helper_container: Container) -> None:
@@ -203,7 +271,7 @@ class VPNManager(DockerBaseManager):
         # add IP route line to the algorithm container network
         cmd = f"ip route replace default via {vpn_local_ip}"
         self.docker.containers.run(
-            image='alpine',
+            image=self.alpine_image,
             network=network,
             cap_add='NET_ADMIN',
             command=cmd,
@@ -211,7 +279,7 @@ class VPNManager(DockerBaseManager):
         )
 
     def _forward_traffic_to_algorithm(self, algo_helper_container: Container,
-                                      algo_image_name: str) -> int:
+                                      algo_image_name: str) -> List[Dict]:
         """
         Forward incoming traffic from the VPN client container to the
         algorithm container
@@ -225,11 +293,20 @@ class VPNManager(DockerBaseManager):
 
         Returns
         -------
-        int
-            Port on the VPN client that forwards traffic to the algo container
+        List[Dict] or None
+            Description of each port on the VPN client that forwards traffic to
+            the algo container. None if VPN is not set up.
         """
         if not self.has_vpn:
             return None  # no port assigned if no VPN is available
+
+        # Get IP Address of the algorithm container
+        algo_helper_container.reload()  # update attributes
+        algo_ip = self.get_isolated_netw_ip(algo_helper_container)
+
+        # Set ports at which algorithm containers receive traffic
+        ports = self._find_exposed_ports(algo_image_name)
+
         # Find ports on VPN container that are already occupied
         cmd = (
             'sh -c '
@@ -239,30 +316,45 @@ class VPNManager(DockerBaseManager):
         occupied_ports = occupied_ports.output.decode('utf-8')
         occupied_ports = occupied_ports.split('\n')
         occupied_ports = \
-            [int(port) for port in occupied_ports if port is not '']
+            [int(port) for port in occupied_ports if port != '']
 
         # take first available port
         vpn_client_port_options = set(FREE_PORT_RANGE) - set(occupied_ports)
-        vpn_client_port = next(iter(vpn_client_port_options))
-
-        # Get IP Address of the algorithm container
-        algo_helper_container.reload()  # update attributes
-        algo_ip = self.get_isolated_netw_ip(algo_helper_container)
-
-        # Set port at which algorithm containers receive traffic
-        algorithm_port = self._find_exposed_port(algo_image_name)
+        for port in ports:
+            port['port'] = vpn_client_port_options.pop()
 
         # Set up forwarding VPN traffic to algorithm container
-        command = (
-            'sh -c '
-            '"iptables -t nat -A PREROUTING -i tun0 -p tcp '
-            f'--dport {vpn_client_port} -j DNAT '
-            f'--to {algo_ip}:{algorithm_port}"'
-        )
+        command = 'sh -c "'
+        for port in ports:
+            command += (
+                'iptables -t nat -A PREROUTING -i tun0 -p tcp '
+                f'--dport {port["port"]} -j DNAT '
+                f'--to {algo_ip}:{port["algo_port"]};'
+            )
+            # remove the algorithm ports from the dictionaries as these are no
+            # longer necessary
+            del port['algo_port']
+        command += '"'
         self.vpn_client_container.exec_run(command)
-        return vpn_client_port
 
-    def _find_exposed_port(self, image: str) -> str:
+        return ports
+
+    def _vpn_in_right_subnet(self) -> bool:
+        """
+        Check if the VPN connection is part of the subnet defined in the node
+        configuration.
+
+        Returns
+        -------
+        bool
+            Whether the VPN IP address is part of the subnet or not
+        """
+        vpn_ip = self.get_vpn_ip()
+        return (
+            ipaddress.ip_address(vpn_ip) in ipaddress.ip_network(self.subnet)
+        )
+
+    def _find_exposed_ports(self, image: str) -> List[Dict]:
         """
         Find which ports were exposed via the EXPOSE keyword in the dockerfile
         of the algorithm image. This port will be used for VPN traffic. If no
@@ -275,33 +367,50 @@ class VPNManager(DockerBaseManager):
 
         Returns
         -------
-        str:
-            Port number to forward VPN traffic to (as str)
+        List[Dict]:
+            List of ports forward VPN traffic to. For each port, a dictionary
+            containing port number and label is given
         """
-        n2n_image = self.docker.images.get(image)
-        port = DEFAULT_ALGO_VPN_PORT
+        default_ports = [{'algo_port': DEFAULT_ALGO_VPN_PORT, 'label': None}]
+        algo_image = self.docker.images.get(image)
 
         exposed_ports = []
         try:
-            exposed_ports = n2n_image.attrs['Config']['ExposedPorts']
+            exposed_ports = algo_image.attrs['Config']['ExposedPorts']
         except KeyError:
-            return port  # No exposed ports defined, use default
+            return default_ports
 
-        if len(exposed_ports) == 1:
-            port = list(exposed_ports)[0]
+        # find any labels defined in the docker image
+        labels = {}
+        try:
+            labels = algo_image.attrs['Config']['Labels']
+        except KeyError:
+            pass  # No labels found, ignore
+
+        ports = []
+        for port in exposed_ports:
             port = port[0:port.find('/')]
             try:
                 int(port)
             except ValueError:
                 self.log.warn("Could not parse port specified in algorithm "
-                              f"docker image {image}: {port}. Using default "
-                              f"port {DEFAULT_ALGO_VPN_PORT}")
-                return DEFAULT_ALGO_VPN_PORT
-        elif len(exposed_ports) > 1:
-            self.log.warn("More than 1 port exposed in docker image. "
-                          f"Using default port {port}.")
-        # else: no exposed port specified, return default
-        return port
+                              f"docker image {image}: {port}")
+            # get port label: this should be defined as 'p1234' for port 1234
+            label = None
+            if labels:
+                label = labels.get('p' + port)
+            if not label:
+                self.log.warn(f"No label defined in image for port {port}. "
+                              "Algorithm will not be able to find the port "
+                              "using the label!")
+            ports.append({'algo_port': port, 'label': label})
+
+        if not ports:
+            self.log.warn(
+                "None of the ports in the algorithm image could be parsed. "
+                f"Using default port {DEFAULT_ALGO_VPN_PORT} instead")
+
+        return ports if ports else default_ports
 
     def _find_isolated_bridge(self) -> str:
         """
@@ -322,7 +431,7 @@ class VPNManager(DockerBaseManager):
 
         # Get network config from host namespace
         host_interfaces = self.docker.containers.run(
-            image=NETWORK_CONFIG_IMAGE,
+            image=self.network_config_image,
             network='host',
             command=['ip', '--json', 'addr'],
             remove=True
@@ -380,33 +489,26 @@ class VPNManager(DockerBaseManager):
         else:
             return False
 
-    def _configure_host_network(self, isolated_bridge: str) -> None:
+    def _configure_host_network(self) -> None:
         """
         By default the internal bridge networks are configured to prohibit
         packet forwarding between networks. Create an exception to this rule
         for forwarding traffic between the bridge and vpn network.
-
-        Parameters
-        ----------
-        vpn_subnet: string
-            Subnet of allowed VPN IP addresses
-        isolated_bridge: string
-            Name of the network interface in the host namespace
         """
         self.log.debug("Configuring host network exceptions for VPN")
         # The following command inserts rules that traffic from the VPN subnet
         # will be accepted into the isolated network
         command = (
             'sh -c "'
-            f'iptables -I DOCKER-USER 1 -d {self.subnet} -i {isolated_bridge} '
-            '-j ACCEPT; '
-            f'iptables -I DOCKER-USER 1 -s {self.subnet} -o {isolated_bridge} '
-            '-j ACCEPT; '
+            f'iptables -I DOCKER-USER 1 -d {self.subnet} '
+            f'-i {self.isolated_bridge} -j ACCEPT; '
+            f'iptables -I DOCKER-USER 1 -s {self.subnet} '
+            f'-o {self.isolated_bridge} -j ACCEPT; '
             '"'
         )
 
         self.docker.containers.run(
-            image=NETWORK_CONFIG_IMAGE,
+            image=self.network_config_image,
             network='host',
             cap_add='NET_ADMIN',
             command=command,
@@ -425,3 +527,16 @@ class VPNManager(DockerBaseManager):
         if isinstance(if_json, list):
             if_json = if_json[-1]
         return int(if_json['link_index'])
+
+    def _is_ipv4_subnet(self, subnet: str) -> bool:
+        """
+        Validate if subnet has format '12.34.56.78/16'
+        """
+        parts = subnet.split('/')
+        if len(parts) != 2:
+            return False
+        if not parts[1].isdigit() or int(parts[1]) > 32:
+            return False
+        octets = parts[0].split(".")
+        return len(octets) == 4 and \
+            all(o.isdigit() and 0 <= int(o) < 256 for o in octets)
